#include <CoreFoundation/CoreFoundation.h>
#import <CommonCrypto/CommonCrypto.h> // For CC_MD5
#include <dlfcn.h>

/**
  * A simple program demonstrating the use of iOS's MobileGestalt
  *
  * Jonathan Levin, 05/23/2015 - http://NewOSXBook.com/
  *
  * Copy at will, but credit or feedback would be appreciated
  * if you find this helpful..
  *
  *
  */


// Confound this CoreFoundation...
char *_CFNumberToString (CFNumberRef 	Num)
{
	char *num = malloc(128); //CFNumberGetByteSize(Num));

	// Should check return value. Meh
	CFNumberGetValue(Num,  // CFNumberRef number
			 CFNumberGetType(Num) , // CFNumberType theType
			 num); // void *valuePtr);

	char *returned = calloc (128,1); // MOOOORE than enough
	switch (CFNumberGetType(Num))
	{
	   case kCFNumberSInt8Type: //  = 1,
		sprintf (returned, "%d", num[0]);
		break;
	   case kCFNumberSInt16Type:  // = 2,
		sprintf (returned, "%d", *((short *) num));
		break;
	   case kCFNumberSInt32Type: // =3,
		sprintf (returned, "%d", *((int32_t *) num));
		break;
	   case kCFNumberSInt64Type: // = 4,
		sprintf (returned, "%lld", *((int64_t *) num));
		break;
	   default:
		return ("?");

#if 0
    /* @TODO, at some point.. Don't get too many values of these types in MG
    kCFNumberFloat32Type = 5,
    kCFNumberFloat64Type = 6,   /* 64-bit IEEE 754 */
    /* Basic C types */
    kCFNumberCharType = 7,
    kCFNumberShortType = 8,
    kCFNumberIntType = 9,
    kCFNumberLongType = 10,
    kCFNumberLongLongType = 11,
    kCFNumberFloatType = 12,
    kCFNumberDoubleType = 13,

#endif
	}
		return (returned);
		
} //CFNumberToString 

// Hold the function pointer to MGCopyAnswer() here.
// You can also define MGCopyAnswer instead of using dlopen(3),
// But this would require -l libMobileGestalt -L .../usr/lib
//
CFTypeRef (*MGCopyAnswer_func)(CFStringRef question);

const char *gestaltAnswer (char *Question)
{


        if (!MGCopyAnswer_func)
        {
                void *gestaltLib = dlopen ("/usr/lib/libMobileGestalt.dylib", RTLD_LAZY);

                if (!gestaltLib)  { /* error */ return NULL;}              

                // Found library, now find symbol
                MGCopyAnswer_func = dlsym (gestaltLib, "MGCopyAnswer");

                if (!MGCopyAnswer_func)  { /* error */ return NULL;}              

        } // end if !MGCopy..

        // Still here, so gestalt is at our disposal!
        

        CFStringRef     q = CFStringCreateWithCString(NULL, 
                                                      Question, 
                                                      kCFStringEncodingASCII);

        CFTypeRef answer = MGCopyAnswer_func(q);
        
        if (!answer)
        {
		// Note that lack of an answer here doesn't necessarily mean that
		// the key was not found - it could also imply a permission problem
		// (q.v. MobileGestalt's "protected keys"
		return(NULL);
        }

	const char *out = NULL;

	// CopyAnswer returns a CF object. Which, we have to figure out dynamically
	// (although you can use one of Get[Bool/Float/SInt[32/64]] variants as well)
        CFTypeID typeId = CFGetTypeID (answer);

	if (typeId == CFStringGetTypeID())
	{
                   out = CFStringGetCStringPtr(answer, kCFStringEncodingASCII);
	}

	if (typeId == CFBooleanGetTypeID())
	{
		   out = (CFBooleanGetValue(answer) ? "true" : "false");
	}
	if (typeId == CFArrayGetTypeID())
	{	
		//@TODO (my plane is landing in a bit :-)
		out = "Array";
	}
	if (typeId == CFDictionaryGetTypeID())
	{
	    CFDataRef xml = CFPropertyListCreateXMLData(kCFAllocatorDefault,
                                                (CFPropertyListRef)answer);
   	     if (xml) { out =  (const char *) CFDataGetBytePtr(xml); }
	     // should really CFRelease(xml);


	} // dict
	if (typeId == CFNumberGetTypeID())
	{
	
		
		out = _CFNumberToString (answer);

	}
	if (typeId == CFDataGetTypeID())
	{
		CFIndex len = CFDataGetLength(answer);
		const UInt8 *data = CFDataGetBytePtr(answer);

		// We want to dump this to hex:
		char *returned = (char *) calloc (len * 3,1);
		
		int i = 0;
		for (i = 0 ; i < len; i++)
		{
		   sprintf(returned + strlen(returned), "%02X ", data[i]);
		}
		
		return (returned);
	}
		   

	return (out);


}


//
// Base64 support
//
char table[] = 
{ 
 'A', 'B', 'C', 'D','E','F','G','H','I', 'J', 'K', 'L', 'M', 
 'N', 'O', 'P', 'Q','R','S','T','U','V', 'W', 'X', 'Y', 'Z',
 'a', 'b', 'c', 'd','e','f','g','h','i', 'j', 'k', 'l', 'm', 
 'n', 'o', 'p', 'q','r','s','t','u','v', 'w', 'x', 'y', 'z',
 '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
 '+',
 '/'

};

char *base64_encode (unsigned char Blob[], uint32_t BlobLen)
{
  // I'd rather create my own, than use NSData and objective-C..
  // Note this won't return padding (=, ==), but since the MG keys aren't padded
  // this is just fine.

 int i = 0;
 int j = 0;
 char *returned = (char *)calloc (BlobLen * 2,1); // technically * 4/3 should be fine..

 for (i = 0; i < BlobLen * 8; i += 6)
 {
	// Get six bits from bit #i

	unsigned char bits = (Blob[i / 8]);

	// if bit 0 - shift 2
        // if bit 1 - shift 1, ignore first one
	// if bit 2 - no shift, just ignore first two
        // if bit 3 - shift left 1, add 1 from next byte..
        // if bit 4 - shift left 2, add 2 from next byte..
        // if bit 6 - shift left 4, add 4 from next byte..
	// if bit 7 - shift left 5, add 5 from next byte..
	// always look at six least significant bits..

	switch (i % 8)
	{
		case 0:
			bits = bits >> 2;
			break;
		case 1:
			bits = (bits & 0x7f) >> 1;
			break;
		case 2: 
			bits = (bits &0x3f);
			break;
		default:
			bits = (bits << ((i % 8) - 2 ));

			bits += ((Blob[i /8 +1]) >> (8 - (i %8) +2));

			bits &= 0x3f;

	}
	
	returned[j++] = table[bits];

 } // end i
 

  return (returned);
 
} // base64

char *encodeKey (char *Key)
{
    // Encode a key, using MobileGestalt's obfuscation.
    // Obfuscated value is a base64 encoding (minus '=' padding)
    // of the string "MGCopyAnswer" and the Key, hashed by MD5

    char obfuscateThis[1024]; // more than enough

    sprintf(obfuscateThis, "%s%s", "MGCopyAnswer", Key);

    unsigned char md5[17];
    CC_MD5(obfuscateThis, strlen(obfuscateThis), md5);
    md5[16] = '\0';

    return (base64_encode (md5, 16));
   
   
} // encodeKey

int main(int argc, char **argv)
{


	if (argc < 2) { fprintf (stderr, "Usage: %s question\n", argv[0]); exit(1);}


	const char *ans = gestaltAnswer(argv[1]);
  	char *enc;

	// Quick check if key is already encoded - if len = 22
	// Note this is not always right, since a key might be 22 chars exactly..
	// So it's better to check for the presence of numbers, which would indicate
  	// base64..

	if (strlen (argv[1]) == 22)  enc = argv[1];
	else  enc= encodeKey (argv[1]);

	if (ans)
	{
	printf ("%s (%s): %s\n", argv[1], enc, gestaltAnswer(argv[1]));
	}
	else
	{
	 printf ("%s (%s): Not found (or possibly permission denied)\n",  argv[1], enc);
	}
	return (0);
} // main
